from movement_primitives.promp import ProMP

import rclpy
from rclpy.node import Node
from rclpy.action import ActionClient

from interaction_msgs.srv import Task
from interaction_msgs.action import TaskExecution

import numpy as np
import asyncio
import threading


class PrompNode(Node):
    def __init__(self, event_loop):
        super().__init__('policy_promp_node')
        self.event_loop = event_loop

        self.task_srv = self.create_service(Task, 'interaction/task_srv', self.task_srv_callback)
        self.execution_action = ActionClient(self, TaskExecution, 'interaction/task_execution')

        # Parameters
        self.timeout_buffer = self.declare_parameter('timeout_buffer', 5.0).get_parameter_value().double_value

    def task_srv_callback(self, request, response):
        coroutine = self.handle_request_async(request, response)

        future = asyncio.run_coroutine_threadsafe(coroutine, self.event_loop)

        try:
            response = future.result(timeout=request.time_array[-1] + self.timeout_buffer)
        except asyncio.TimeoutError:
            self.get_logger().error('Task timed out!')

        return response

    async def handle_request_async(self, request, response):
        if len(request.x_next) != request.nr_dim * request.nr_parameter:
            self.get_logger().error('x_next size does not math to nr_dim * nr_parameter')
            return response

        time_array = np.array(request.time_array, dtype=np.float32)
        weights = np.array(request.x_next, dtype=np.float32).reshape(request.nr_dim, request.nr_parameter)

        promp = ProMP(n_dims=request.nr_dim, n_weights_per_dim=request.nr_parameter)
        trajectory = promp.trajectory_from_weights(time_array, weights)

        goal = TaskExecution.Goal()
        goal.trajectory = trajectory
        await self.execution_action.wait_for_server()
        send_goal_future = self.execution_action.send_goal_async(goal)
        goal_handle = await send_goal_future

        if not goal_handle.accepted:
            self.get_logger().error('Goal rejected!')
            return response

        self.get_logger().info('Goal accepted!')
        result_handle = await goal_handle.get_result_async()

        response.x_observed = request.x_next
        response.reward = result_handle.result.reward

        return response


def run_asyncio_loop():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_forever()
    return loop  # Ensure you return the loop reference


def main():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    threading.Thread(target=loop.run_forever, daemon=True).start()

    # Initialize ROS and pass the event loop to the node
    rclpy.init()
    node = PrompNode(loop)

    # Now, the ROS 2 node has the correct event loop reference for asyncio operations
    rclpy.spin(node)

    node.destroy_node()
    rclpy.shutdown()
    loop.call_soon_threadsafe(loop.stop)


if __name__ == '__main__':
    main()
